/*
 * Copyright (c) 2005
 *	Deutsches Zentrum fuer Luft- und Raumfahrt DLR.
 *	All rights reserved.
 *
 * Author: Harti Brandt <hartmut.brandt@dlr.de>
 *
 * Redistribution of this software and documentation and use in source and
 * binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 *
 * 1. Redistributions of source code or documentation must retain the above
 *    copyright notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE AND DOCUMENTATION IS PROVIDED BY FRAUNHOFER FOKUS
 * AND ITS CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * FRAUNHOFER FOKUS OR ITS CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ip.h 582 2006-12-12 14:56:00Z brandt_h $
 */

#ifndef ip_h_1115889008
#define ip_h_1115889008

#include <sys/types.h>
#include <sys/socket.h>
#include "histogram.h"

/* global debug flag */
extern u_int iplib_debug;

/**
 * Structure to describe the IP part of a connection
 */
struct mod_ip {
	struct {
		u_char	a[4];		/*!< IP address */
	}		addr;
	int		keepalive : 1;	/*!< SO_KEEPALIVE */
	int		dontfrag : 1;	/*!< IP_DONTFRAG */
	u_int		tos : 8;	/*!< IP_TOS */
	u_int		ttl : 8;	/*!< IP_TTL */
	uint32_t	sndbuf;		/*!< SO_SNDBUF */
	uint32_t	rcvbuf;		/*!< SO_RCVBUF */
};

/**
 * Structure to describe a TCP socket. The IP config is supposed to be
 * at the start of this structure.
 */
struct mod_tcp {
	struct mod_ip	ip;		/*!< the IP part */
	uint16_t	port;		/*!< TCP port number */
	uint16_t	mss;		/*!< TCP_MAXSEG */
	int		nodelay : 1;	/*!< TCP_NODELAY */
	int		nopush : 1;	/*!< TCP_NOPUSH */
	int		noopt : 1;	/*!< TCP_NOOPT */
};

/**
 * Structure to describe a UDP socket. The IP config is supposed to be
 * at the start of this structure.
 */
struct mod_udp {
	struct mod_ip	ip;		/*!< the IP part */
	uint16_t	port;		/*!< UDP port number */
	uint32_t	long_thresh;	/*!< long error threshold */
};

/**
 * Wire format for UDP.
 *
 * The format consists of a library header, optionally followed by an
 * application specific header. The CRC is computed with the CRC field set
 * to 0.
 */
struct udp_packet {
	uint64_t	tstamp;		/* sending time */
	uint16_t	len;		/* size of packets in bytes */
	uint16_t	seqno;		/* sequence number */
	uint32_t	crc;		/* checksum */
};
#define	UDP_MAXLEN	65536

/**
 * UDP receive statistics
 */
struct udp_recv_stats {
	uint64_t	recv_failed;	/* recv() returned an error */
	uint64_t	short_packet;	/* length < sizeof(udp_packet) */
	uint64_t	long_packet;	/* packet truncated */
	uint64_t	len_mismatch;	/* length != udp_packet.len */
	uint64_t	bad_crc;	/* CRC bad */
	uint64_t	short_losses;	/* short runs of loss */
	uint64_t	long_losses;	/* long runs of loss */
	uint64_t	packets_lost;	/* number of lost packets */
	uint64_t	packets_late;	/* number of reordered packets */
	uint64_t	packets_rcvd;	/* number of received good packets */
	uint64_t	bytes_rcvd;	/* number of received good bytes */
	uint64_t	n_iat;		/* packets counted for IAT */
	double		s_iat;		/* sum of IATs */
	double		q_iat;		/* sum of square of delay */
	HG_DECLARE(SIM_HIST_LEN, u64) iat; /* interarrival time histogram */
	uint64_t	n_delay;	/* packets counted for delay */
	double		s_delay;	/* sum of delay */
	double		q_delay;	/* sum of square of delay */
	HG_DECLARE(SIM_HIST_LEN, u64) delay; /* delay histogram */
	uint64_t	bad_tstamps;	/* bad timestamp count */
};

/**
 * UDP receiver configuration.
 */
struct udp_recv_config {
	uint32_t	long_thresh;	/* threshold for long losses */
	struct udp_packet *packet;	/* packet buffer */
	size_t		minlen;		/* minimum packet length */
	size_t		maxlen;		/* maximum length */
	uint64_t	tstamp;		/* receive timestamp */
	uint16_t	next_packet;	/* next expected seqno */
	u_int		loss_run;	/* for testing */
};

/**
 * UDP receive stats config
 */
struct udp_recv_stats_config {
	struct hg_u64_config	iat;
	struct hg_u64_config	delay;
};

/**
 * UDP sender statistics
 */
struct udp_send_stats {
	uint64_t	packets_sent;	/* number of successful packets */
	uint64_t	bytes_sent;	/* number of bytes in these packets */
	uint64_t	send_failed;	/* total send failures */
	uint64_t	conn_refused;	/* ECONNREFUSED in these */
	uint64_t	short_sends;	/* sent < to-send */
};

/**
 * UDP sender configuration
 */
struct udp_send_config {
	struct udp_packet *packet;	/* packet buffer */
	uint16_t	next_packet;	/* next sequence number */
	size_t		header_len;	/* total length of headers */
};

/**
 * Wire format for TCP.
 *
 * The format consists of a library header, optionally followed by an
 * application specific header. This is followed by a payload that consists
 * of chunks. Each of these chunks has its own timestamp and CRC.
 * The last chunk may be shorter than the others. If even the chunk header
 * doesn't fit - just send junk.
 */
struct tcp_packet {
	uint64_t	tstamp;		/* sending time */
	uint32_t	len;		/* size of packets in bytes */
	uint32_t	crc;		/* header only */
};
struct tcp_data {
	uint64_t	tstamp;		/* sending time */
	uint32_t	crc;		/* entire data packet */
#define	TCP_CHUNK_HDRLEN	12
#define	TCP_CHUNK_LEN		256
	u_char		data[TCP_CHUNK_LEN - TCP_CHUNK_HDRLEN];
};

/**
 * TCP statistics
 */
struct tcp_stats {
	uint64_t	n_connects;	/* A successful connects */
	double		s_connects;	/* A sum of establish times */
	double		q_connects;	/* A sum of square of the same */
	uint64_t	t_connects;	/* A failed connects */

	uint64_t	n_conn_ok;	/* A number of successful finished */
	double		s_conn_ok;	/* A sum of connection times */
	double		q_conn_ok;	/* A sum of square of connection ts */

	uint64_t	bytes_rcvd;	/* bytes received */
	uint64_t	bytes_sent;	/* bytes sent */

	uint64_t	conn_refused;	/* + ECONNREFUSED errors */
	uint64_t	host_unreach;	/* + host unrechable */
	uint64_t	conn_reset;	/* + RESET received */
	uint64_t	conn_err;	/* + other connect error */
	uint64_t	poll_failed;	/* + poll_register call failed */
	uint64_t	malloc_failed;	/* + malloc call failed */
	uint64_t	send_failed;	/* S send() call returns -1 */
	uint64_t	listen_failed;	/* L listen() call returns -1 */
	uint64_t	accept_failed;	/* L accept() call returns -1 */
	uint64_t	recv_failed;	/* R recv() call returns -1 */
	uint64_t	unexp_eof;	/* R recv() returns 0 */
	uint64_t	bad_crc;	/* R CRC error */

	uint16_t	neg_mss;	/* + negotiated MSS */
};

/**
 * TCP connect configuration
 */
struct tcp_sock {
	/* to be initiated by the caller */
	struct mod_tcp	*tcp;		/* TCP configuration */
	void		*arg;		/* user arg */
	void		(*finish_connect)(struct tcp_sock *, int);
	void		(*finish_send)(struct tcp_sock *, int);
	void		(*finish_recv)(struct tcp_sock *, ssize_t);
	void		(*newconn)(struct tcp_sock *, int);

	struct tcp_stats *stats;
	struct tcp_stats *gstats;

	struct tcp_packet *header;	/* packet header */
	size_t		header_len;	/* length of header */

	int		must_recv : 1;	/* TRUE: EOF is illegal */

	/* public */
	uint64_t	conn_start;	/* starting timestamp */

	/* internal */
	int		sock;		/* the socket */
	int		h_in;		/* poll handle */
	int		h_out;		/* poll handle */

	struct tcp_data	send_data;	/* data */
	u_char		*send_ptr;	/* either header or &data */
	size_t		bytes_sent;	/* number of bytes sent from current */
	size_t		to_send;	/* total number to send from current */
	size_t		total_sent;	/* current packet sent */

	struct tcp_data	recv_data;	/* data */
	u_char		*recv_ptr;	/* either header or &data */
	size_t		bytes_rcvd;	/* number of bytes rcvd from current */
	size_t		to_recv;	/* number to rcvd from current */
	size_t		total_rcvd;	/* total for this packet */
};

/* start connecting to the given TCP socket. */
int tcp_connect(struct tcp_sock *);

/* Start listening to the given TCP socket. */
int tcp_listen(struct tcp_sock *);

/* Create new connection with already open socket. */
int tcp_accept(struct tcp_sock *, int);

/* Stop anything on that TCP and free all resources. */
void tcp_stop(struct tcp_sock *, int);

/* send TCP packet */
int tcp_send(struct tcp_sock *, size_t);
void tcp_send_stop(struct tcp_sock *);

/* receive TCP packet */
int tcp_recv(struct tcp_sock *);
void tcp_recv_stop(struct tcp_sock *);

/* reset TCP statistics */
void tcp_stats_reset(struct tcp_stats *);

/* create and connect a UDP socket */
int udp_connect(const struct mod_udp *);

/* create and bind a UDP socket */
int udp_listen(const struct mod_udp *);

/* send a UDP packet */
int udp_send(int, struct udp_send_config *, struct udp_send_stats *,
    struct udp_send_stats *, size_t);

/* reset UDP send statistics */
void udp_send_stats_reset(struct udp_send_stats *);

/* receive UDP packet */
int udp_recv(int, struct udp_recv_config *, struct udp_recv_stats *,
    struct udp_recv_stats *);

/* reset UDP receive statistics */
void udp_recv_stats_reset(struct udp_recv_stats *);

/* hexdump data */
void dump_data_block(const void *, size_t);

#endif
