/*-
 * Copyright (c) 2004-2005
 *	German Aerospace Center (DLR).
 * 	All rights reserved.
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: libconf.h 278 2006-03-27 14:28:56Z brandt_h $
 */
#ifndef LIBCONF_H_1116321538
#define LIBCONF_H_1116321538

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>
#include <stdarg.h>
#include <stddef.h>

#define	CONF_MAXARG	100

struct conf_conf {
	void	*data;		/* user data */
};

/**
 * Description of the arguments for a command. These arguments consist
 * of options (with optional arguments) and (possibly) optional arguments.
 */
struct conf_opt {
	const char	*str;		/* the option string */
	const char	*argstr;	/* short argument description */
	int		(*func)(const char *, int, char **,
			    const struct conf_opt *, void *);
	ptrdiff_t	optarg;		/* argument offset */
	const void	*argarg;	/* argument to parsing function */
};

struct conf_cmdline {
	size_t		argsize;	/* size of the argument struct */
	const void	*def;		/* default data */
	int		(*func)(struct conf_conf *, void *); /* handler */
	u_int		options;	/* number of options */
	u_int		reqargs;	/* number of required arguments */
	struct conf_opt	args[];		/* pointer to option description */
};

int conf_opt_simple(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_uint(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_int(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_uint16(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_int16(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_uint32(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_int32(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_uint64(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_int64(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_flag(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_string(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_double(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_uint32list(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_strlist(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_doublelist(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_ipv4host(const char *, int, char **,
    const struct conf_opt *, void *);
int conf_opt_enum(const char *, int, char **,
    const struct conf_opt *, void *);

/* Allocate memory and put it onto the to-free list. */
void *conf_opt_malloc(size_t);

/* Print an error from an option parsing function */
void conf_opt_error(const struct conf_opt *, const char *, ...);

#define	CONF_OPT_SET_FLAG(RET, CONF, VAR) do {				\
	(VAR) = ((char *)(RET) + (CONF)->optarg);			\
	*(int *)(VAR) = 1;						\
	(VAR) = ((char *)(VAR) + sizeof(int));				\
    } while (0)

#define	CONF_OPT_SET_VALUE(VAR, TYPE, VAL) do {				\
	*(TYPE *)(VAR) = (TYPE)(VAL);					\
	(VAR) = ((char *)(VAR) + sizeof(TYPE));				\
    } while (0)

#define	CONF_OPT_MOVE_ADDR(VAR, TYPE) do {				\
	(VAR) = ((char *)(VAR) + sizeof(TYPE));				\
    } while (0)

/**
 * For conf_opt_string() this structure allows to limit the string to
 * certain sizes.
 */
struct lisa_opt_string {
	size_t	minlen;
	size_t	maxlen;
};

/**
 * This structure describes one level of command. The field string is
 * a pointer to the string that the user enters. If the command has
 * sub-commands (a next lower level) then sub must point the command table
 * for the sub-commands. If the command can execute something directly,
 * func is the handler for it.
 *
 * All commands at a given level must be described by an array of such structs.
 * The array is terminated by a struct with a NULL string.
 */
struct conf_cmdtab {
	const char	*string;
	const struct conf_cmdtab *sub;
	const struct conf_cmdline *cmd;
};

/* parse a command line and call the handler */
int conf_parse_cmdline(struct conf_conf *, int, char **,
    const struct conf_cmdline *);

/**
 * Configuration for a module.
 */
struct conf_amodule_1 {
	const char		*comment;
	const struct conf_cmdtab *cmd;
};

#ifdef CONF_BUILTIN
#define	CONF_DEF_MODULE(COMMENT, NAME)					\
extern const struct conf_amodule_1 conf_amodule_1;			\
void NAME##_register(void);						\
void									\
NAME##_register(void)							\
{									\
	conf_register_module_1(&conf_amodule_1);			\
}									\
static const struct conf_amodule_1 conf_amodule_1 = { COMMENT, NAME##_main }
#else
#define	CONF_DEF_MODULE(COMMENT, NAME)					\
const struct conf_amodule_1 conf_amodule_1 = { COMMENT, NAME##_main }
#endif

/* verbosity level */
extern int conf_verbose;

/* program name */
extern const char *conf_progname;

/* currently executed command */
extern char *conf_current;

/* notitle option */
extern int conf_notitle;

/* initialize command table */
struct conf_conf *conf_init(const char *, const struct conf_cmdtab *);

/* for compiled-in modules */
void conf_register_module_1(struct conf_conf *,
    const struct conf_amodule_1 *);

/* print a message if we are verbose */
void	conf_verb(const char *, ...) __printflike(1, 2);

/* print heading */
void	conf_heading(const char *, ...) __printflike(1, 2);

/* before starting output */
void	conf_heading_init(void);

/* call help directly */
int conf_help_main(struct conf_conf *, int, char **);

/* print an error */
void	conf_error_start(const char *, ...);
void	conf_error_vstart(const char *, va_list);
void	conf_error_end(void);
void	conf_error_cont(const char *, ...);
void	conf_error_vcont(const char *, va_list);
void	conf_error(const char *, ...);

/* stringify an enumerated value */
struct conf_enum {
	u_int	value;
	const char *str;
};

/* print an enum */
const char *conf_enum_print(u_int, const struct conf_enum *, char *);

/* find a string in an enum list */
int conf_enum_parse(u_int *, const struct conf_enum *, const char *);

int conf_substr(const char *, const char *);

/* run a command */
int conf_run(struct conf_conf *, int, char **);

/* run a command line */
int conf_run_line(struct conf_conf *, const char *);

/* source a file */
int conf_source_file(struct conf_conf *, const char *);

/* get the path to the config file directories */
const char *conf_get_confpath(struct conf_conf *, const char *);

/* set the configuration file path */
int conf_set_confpath(struct conf_conf *, const char *, const char *);

/* User callable function to add nodes. */
int conf_node_add(struct conf_conf *, int, const char **,
    const struct conf_cmdtab *);

int conf_parse_uint32(const char *, uint32_t *, int);

/* replace variables in a string */
char *conf_replace_vars(struct conf_conf *, const char *);

#endif /* LIBCONF_H_1116321538 */
